"""
Trizna, D. and Demetrio, L. and Battista, B. and Fabio, R. (2024)
Nebula: Self-Attention for Dynamic Malware Analysis
IEEE Transactions on Information Forensics and Security (TIFS)
Paper: https://ieeexplore.ieee.org/document/10551436
Source: https://github.com/dtrizna/nebula/blob/main/nebula/preprocessing/normalization.py
"""

import re
import pandas as pd
from .constants import VARIABLE_MAP


def normalize_string_hash(string: str) -> str:
    if not isinstance(string, str):
        return string
    string = re.sub(r'[0-9a-fA-F]{64}', "<sha256>", string)
    string = re.sub(r'[0-9a-fA-F]{40}', "<sha1>", string)
    string = re.sub(r'[0-9a-fA-F]{32}', "<md5>", string)
    return string


def normalize_string_domain(
    string: str,
    domain_regex: str = r"\b([a-zA-Z0-9][a-zA-Z0-9-]{0,61}[a-zA-Z0-9]\.)+",
    top_level_domains_regex: str = r"(com|net|ai|us|uk|cz|gov)\b",
    placeholder: str = "<domain>"
) -> str:
    if not isinstance(string, str):
        return string
    full_regex = domain_regex + top_level_domains_regex
    return re.sub(full_regex, placeholder, string)


def normalize_string_ip(string: str) -> str:
    if not isinstance(string, str):
        return string
    string = re.sub(r'127\.\d{1,3}\.\d{1,3}\.\d{1,3}', "<lopIP>", string)
    string = re.sub(r'169\.254\.169\.254', "<imds>", string)
    string = re.sub(
        r'(10(\.\d{1,3}){3}|(172\.(1[6-9]|2[0-9]|3[0-1])|192\.168)(\.\d{1,3}){2})',
        "<prvIP>",
        string
    )
    string = re.sub(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', "<pubIP>", string)
    # Exclude IPv6 for now
    # string = re.sub(r'([0-9a-fA-F]{1,4}:){1,7}:', "<IPv6>", string)
    return string


def normalize_string_path(path: str) -> str:
    """
    Normalizes a file path by substituting drive letters, network hosts, user names, and environment variables.
    """
    if not isinstance(path, str):
        return path

    # Clean up auxiliary strings
    path = path.lower().replace("*raw:", "").replace("*amsiprocess:", "").replace("a script started by ", "").strip()

    # Normalize drive letters
    path = re.sub(r"\w:\\{1,2}", r"<drive>\\", path)

    # Normalize network paths
    path = re.sub(r"[\\]{1,2};((lanmanredirector|webdavredirector)\\;)?\w\:[a-z0-9]{16}", r"\\", path)
    path = re.sub(r"\\\\[\w\d.\-]+\\", r"<net>\\", path)

    # Normalize drive if path starts with single backslash
    path = re.sub(r"^\\([^\\])", r"<drive>\\\1", path)

    # Normalize volume GUID paths
    path = re.sub(r"\\[\.\?]\\volume\{[a-z0-9\-]{36}\}", r"<drive>", path)

    # Normalize non-default users
    default_users = ["administrator", "public", "default"]
    if "users\\" in path:
        if not any(f"users\\{user}\\" in path for user in default_users):
            path = re.sub(r"users\\[^\\]+\\", r"users\\<user>\\", path)

    # Replace environment variables
    for k, v in VARIABLE_MAP.items():
        path = path.replace(k, v)

    return path


def normalize_table_path(df: pd.DataFrame, col: str = "path") -> pd.DataFrame:
    if df.empty:
        return df
    df = df.copy()
    df[col] = df[col].apply(normalize_string_path)
    df[col] = df[col].apply(normalize_string_hash)
    return df


def normalize_table_ip(df: pd.DataFrame, col: str = 'auditd.summary.object.primary') -> pd.DataFrame:
    if df.empty:
        return df
    df = df.copy()
    df[col] = df[col].apply(normalize_string_ip)
    return df



from typing import Tuple

import torch
from random import randint
from maltorch.datasets.dynamic_drs_dataset import DynamicChunkSizeDRSDataset

class DynamicRandomDRSDataset(DynamicChunkSizeDRSDataset):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le, Jordi Planes
    Adversarial Robustness of Deep Learning-based Malware Detectors via (De) Randomized Smoothing
    IEEE Access 2024 - Random Chunks-based (De)Randomized Smoothing
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2 ** 20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 sort_by_size: bool = False,
                 file_percentage: float = 0.05,
                 num_chunks: int = 100,
                 is_training: bool = True,
                 min_chunk_size=500):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            padding_idx=padding_idx,
            min_len=min_len,
            sort_by_size=sort_by_size,
            file_percentage=file_percentage,
            num_chunks=num_chunks,
            is_training=is_training,
            min_chunk_size=min_chunk_size
        )

    def generate_testing_examples(self, batch: list) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        During testing, given an input example we sequentially extract N chunks of bytes. This chunks may overlap with
        one another depending on the amount of chunks you want to extract and the size of each chunk
        """
        if len(batch) == 1:  # Only implemented for batch sizes equals to 1
            # Code here
            x = batch[0][0]
            max_size = x.size()[0]
            chunk_size = int(-(-(max_size * self.file_percentage) // 1))  # Used to round up the float value
            chunk_size = max(self.min_chunk_size, chunk_size) # The chunk size has to be at least equal to the kernel size of the first convolutional layer

            vecs = []
            for _ in range(self.num_chunks):
                curr_idx = randint(0, len(x) - chunk_size)
                vecs.append(x[curr_idx:curr_idx + chunk_size])

            x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
            y = batch[0][1]
            return x, y
        else:
            raise NotImplementedError





"""
Learning the PE Header, Malware Detection with Minimal Domain Knowledge
E. Raff, J. Sylvester, C. Nicholas
https://dl.acm.org/doi/10.1145/3128572.3140442
"""
from torch import nn
from maltorch.zoo.model import EmbeddingModel


class FCNN(EmbeddingModel):

    def __init__(self,
                 embedding_size: int = 16,
                 max_len: int = 512,
                 min_len: int = 512,
                 threshold: float =0.5,
                 padding_idx: int = 256,
                 hidden_dims: tuple = (512, 512, 512, 512),
                 p_embed_dropout: float = 0.20,
                 p_hidden_dropout: float = 0.50):
        super(FCNN, self).__init__(
           name="FCNN", gdrive_id=None, max_len=max_len, min_len=min_len
        )
        self.embedding_1 = nn.Embedding(
           num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )
        self.embedding_1_dropout = nn.Dropout(p_embed_dropout)

        in_dim = max_len * embedding_size
        self.dense_1 = nn.Linear(in_features=in_dim, out_features=hidden_dims[0])
        self.bn1 = nn.BatchNorm1d(hidden_dims[0])
        self.dense_2 = nn.Linear(in_features=hidden_dims[0], out_features=hidden_dims[1])
        self.bn2 = nn.BatchNorm1d(hidden_dims[1])
        self.dense_3 = nn.Linear(in_features=hidden_dims[1], out_features=hidden_dims[2])
        self.bn3 = nn.BatchNorm1d(hidden_dims[2])
        self.dense_4 = nn.Linear(in_features=hidden_dims[2], out_features=hidden_dims[3])
        self.bn4 = nn.BatchNorm1d(hidden_dims[3])

        self.out = nn.Linear(in_features=hidden_dims[3], out_features=1)
        self.dropout = nn.Dropout(p_hidden_dropout)
        self.act = nn.ELU()

        self.embedding_size = (embedding_size,)
        self.max_len = max_len
        self.threshold = threshold
        self.invalid_value = padding_idx
        self.hidden_dims = hidden_dims
        self.p_embed_dropout = p_embed_dropout
        self.p_hidden_dropout = p_hidden_dropout

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def _forward_embed_x(self, x):
        x = self.embedding_1_dropout(x)
        z = x.reshape(x.size(0), -1)

        # FC Blocks
        h1 = self.dropout(self.act(self.bn1(self.dense_1(z))))
        h2 = self.dropout(self.act(self.bn1(self.dense_2(h1))))
        h3 = self.dropout(self.act(self.bn1(self.dense_3(h2))))
        h4 = self.dropout(self.act(self.bn1(self.dense_4(h3))))
        out = self.out(h4)
        return out

    def embedding_matrix(self):
        return self.embedding_1.weight








"""
Marek Krcál, Ondrej Svec, Martin Bálek, and Otakar Jasek.
Deep convolutional malware classifiers can learn from raw executables and labels only.
ICLR 2018. Workshop Track Proceedings, 2018.
https://openreview.net/pdf?id=HkHrmM1PM
"""
from typing import Optional, Union

import torch
import torch.nn.functional as F
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing
from torch import nn

from maltorch.data_processing.e2e_preprocessor import PaddingPreprocessing
from maltorch.zoo.model import EmbeddingModel, BaseEmbeddingPytorchClassifier


class AvastStyleConv(EmbeddingModel):

    DEFAULT_MIN_LENGTH = 10244
    DEFAULT_MAX_LENGTH = 512000

    def __init__(self,
                 embedding_size: int = 8,
                 min_len: int = DEFAULT_MIN_LENGTH,
                 max_len: int = DEFAULT_MAX_LENGTH,
                 threshold: float = 0.5,
                 padding_idx: int = 256,
                 channels: int = 48,
                 window_size: int = 32,
                 stride: int = 4):
        #https://drive.google.com/file/d/1KOB-o-2avfPtsQaGuguRW-b1HDbX8r8v/view?usp=drive_link
        super(AvastStyleConv, self).__init__(
            name="AvastStyleConv", gdrive_id="1KOB-o-2avfPtsQaGuguRW-b1HDbX8r8v", min_len=min_len, max_len=max_len
        )
        self.max_len = max_len
        self.threshold = threshold
        self.invalid_value = padding_idx
        self.channels = channels
        self.window_size = window_size
        self.stride = stride

        self.embedding_1 = nn.Embedding(
            num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )

        self.conv1d_1 = nn.Conv1d(8, self.channels, self.window_size, stride=self.stride, bias=True)
        self.conv1d_2 = nn.Conv1d(self.channels, self.channels * 2, self.window_size, stride=self.stride, bias=True)
        self.pool_1 = nn.MaxPool1d(4)
        self.conv1d_3 = nn.Conv1d(self.channels * 2, self.channels * 3, self.window_size // 2, stride=self.stride * 2,
                                  bias=True)
        self.conv1d_4 = nn.Conv1d(self.channels * 3, self.channels * 4, self.window_size // 2, stride=self.stride * 2,
                                  bias=True)

        self.dense_1 = nn.Linear(self.channels * 4, self.channels * 4)
        self.dense_2 = nn.Linear(self.channels * 4, self.channels * 3)
        self.dense_3 = nn.Linear(self.channels * 3, self.channels * 2)
        self.dense_4 = nn.Linear(self.channels * 2, 1)

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def embedding_matrix(self):
        return self.embedding_1.weight

    def _forward_embed_x(self, x):
        conv1d_1 = torch.relu(self.conv1d_1(x))
        conv1d_2 = torch.relu(self.conv1d_2(conv1d_1))
        pool_1 = self.pool_1(conv1d_2)
        conv1d_3 = torch.relu(self.conv1d_3(pool_1))
        conv1d_4 = torch.relu(self.conv1d_4(conv1d_3))

        global_avg_pooling1d_1 = F.avg_pool1d(
            input=conv1d_4, kernel_size=conv1d_4.size()[2:]
        )
        global_avg_pooling1d_1_flatten = global_avg_pooling1d_1.view(
            global_avg_pooling1d_1.size(0), -1
        )

        dense_1 = self.dense_1(global_avg_pooling1d_1_flatten)
        dense_1_activation = torch.selu(dense_1)
        dense_2 = self.dense_2(dense_1_activation)
        dense_2_activation = torch.selu(dense_2)
        dense_3 = self.dense_3(dense_2_activation)
        dense_3_activation = torch.selu(dense_3)
        dense_4 = self.dense_4(dense_3_activation)
        return dense_4

    @classmethod
    def create_model(
            cls,
            model_path: Optional[str] = None,
            device: str = "cpu",
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
            threshold: Optional[Union[float, None]] = 0.5,
            **kwargs,
    ) -> BaseEmbeddingPytorchClassifier:
        if preprocessing is None:
            preprocessing = PaddingPreprocessing(max_len=AvastStyleConv.DEFAULT_MAX_LENGTH)
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net = net.to(device)  # Explicitly load model to device
        net = net.eval()
        net = BaseEmbeddingPytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
            threshold=threshold,
        )
        return net


import torch
from torch import Tensor

from maltorch.initializers.initializers import ByteBasedInitializer, IdentityInitializer
from maltorch.manipulations.byte_manipulation import ByteManipulation


class GradientFreeWrapperManipulation(ByteManipulation):
    def __init__(
        self,
        inner_manipulation: ByteManipulation,
        domain_constraints=None,
        perturbation_constraints=None,
    ):
        if domain_constraints is None:
            domain_constraints = []
        if perturbation_constraints is None:
            perturbation_constraints = []
        self.inner_manipulation = inner_manipulation
        super().__init__(IdentityInitializer(), domain_constraints, perturbation_constraints)

    def _apply_manipulation(
        self, x: torch.Tensor, delta: torch.Tensor
    ) -> tuple[Tensor, Tensor]:
        x, scaled_delta = self.inner_manipulation(x, delta * 255)
        return x, scaled_delta / 255

    def initialize(self, samples: torch.Tensor):
        self.indexes_to_perturb = []
        modified_sample, delta, indexes = self.inner_manipulation.initializer(samples.data)
        self.indexes_to_perturb = indexes
        samples.data = modified_sample
        return samples, delta / 255



JSON_CLEANUP_SYMBOLS = ['"', "'", ":", ",", "[", "]", "{", "}", "\\", "/"]

SPEAKEASY_RECORD_LIMITS = {"network_events.traffic": 256}

SPEAKEASY_RECORD_FIELDS = [
    'file_access.event',
    'file_access.path',
    'network_events.traffic.server',
    'network_events.traffic.port',
    'registry_access.event',
    'registry_access.path',
    'apis.api_name',
    'apis.args',
    'apis.ret_val',
]

SPEAKEASY_TOKEN_STOPWORDS = ['api_name', 'args', 'ret_val', 'event', 'path', 'open_flags', 'access_flags', 'size', 'server', 'proto', 'port', 'method']

QUO_VADIS_LABELMAP = {
    "clean": 0,
    "backdoor": 1,
    "coinminer": 2,
    "dropper": 3,
    "keylogger": 4,
    "ransomware": 5,
    "rat": 6,
    "trojan": 7
}

# good reference:
# https://docs.microsoft.com/en-us/windows/deployment/usmt/usmt-recognized-environment-variables
VARIABLE_MAP = {
    r"%systemdrive%": r"<drive>", 
    r"%systemroot%": r"<drive>\windows",
    r"%windir%": r"<drive>\windows", 
    r"%allusersprofile%": r"<drive>\programdata",
    r"%programdata%": r"<drive>\programdata",
    r"%programfiles%": r"<drive>\program files",
    r"%programfiles(x86)%": r"<drive>\program files (x86)",
    r"%programw6432%": r"<drive>\program files",
    r"%commonprogramfiles%": r"<drive>\program files\common files",
    r"%commonprogramfiles(x86)%": r"<drive>\program files (x86)\common files",
    r"%commonprogramw6432%": r"<drive>\program files\common files",
    r"%commonfiles%": r"<drive>\program files\common files",
    r"%profiles%": r"<drive>\users",
    r"%public%": r"<drive>\users\public",
    r"%userprofile%": r"<drive>\users\<user>"
}
# more user variables
VARIABLE_MAP.update({
    r"%homepath%": VARIABLE_MAP[r"%userprofile%"],
    r"%downloads%": VARIABLE_MAP[r"%userprofile%"] + r"\downloads",
    r"%desktop%": VARIABLE_MAP[r"%userprofile%"] + r"\desktop",
    r"%favorites%": VARIABLE_MAP[r"%userprofile%"] + r"\favorites",
    r"%documents%": VARIABLE_MAP[r"%userprofile%"] + r"\documents",
    r"%mydocuments%": VARIABLE_MAP[r"%userprofile%"] + r"\documents", # obsolete
    r"%personal%": VARIABLE_MAP[r"%userprofile%"] + r"\documents", # obsolete
    r"%localsettings%": VARIABLE_MAP[r"%userprofile%"] + r"\documents", # obsolete
    r"%mypictures%": VARIABLE_MAP[r"%userprofile%"] + r"\documents\my pictures",
    r"%mymusic%": VARIABLE_MAP[r"%userprofile%"] + r"\documents\my music",
    r"%myvideos%": VARIABLE_MAP[r"%userprofile%"] + r"\documents\my videos",
    r"%localappdata%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\local",
    r"%appdata%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\roaming",
    r"%usertemp%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\local\temp",
    r"%temp%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\local\temp",
    r"%tmp%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\local\temp",
    r"%cache%": VARIABLE_MAP[r"%userprofile%"] + r"\appdata\local\microsoft\windows\temporary internet files"
})    


import torch
from secmlt.models.data_processing.data_processing import DataProcessing


class RandomizedAblationPreprocessing(DataProcessing):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le
    Towards a Practical Defense Against Adversarial Attacks on Deep Learning-Based Malware Detectors via Randomized
    Smoothing.
    ESORICS Workshops, SECAI 2024
    """
    def __init__(self, pabl: float = 0.20, num_versions: int = 100, padding_idx: int = 256):
        super().__init__()
        self.pabl = pabl
        self.num_versions = num_versions
        self.padding_idx = padding_idx

    def _process(self, x: torch.Tensor) -> torch.Tensor:
        x = x.squeeze()  # Remove all dimensions equal to 1
        vecs = []
        for i in range(self.num_versions):
            mask = torch.rand(x.shape[0]) <= self.pabl
            mask = mask.to(x.device)
            masked_x = x.masked_fill(mask, self.padding_idx)
            vecs.append(masked_x)
        x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
        return x

    def invert(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        raise NotImplementedError




from abc import ABC, abstractmethod

import torch
from secmlt.optimization.initializer import Initializer


class ByteBasedInitializer(Initializer, ABC):
    def __init__(self, random_init: bool = False):
        self.random_init = random_init

    @staticmethod
    def _pad_samples_same_length(list_of_samples, list_of_deltas, list_of_indexes):
        x = (
            torch.nn.utils.rnn.pad_sequence(list_of_samples, padding_value=256)
            .transpose(0, 1)
            .float()
        )
        delta = (
            torch.nn.utils.rnn.pad_sequence(list_of_deltas, padding_value=256)
            .transpose(0, 1)
            .float()
        )
        indexes = (
            torch.nn.utils.rnn.pad_sequence(list_of_indexes, padding_value=-1)
            .transpose(0, 1)
            .long()
        )
        return x, delta, indexes

    @abstractmethod
    def __call__(self, x: torch.Tensor) -> [torch.Tensor, torch.Tensor, list]: ...


class IdentityInitializer(ByteBasedInitializer):
    def __init__(self, random_init: bool = False):
        super().__init__(random_init=random_init)

    def __call__(self, x: torch.Tensor) -> [torch.Tensor, torch.Tensor, list]:
        delta = torch.rand_like(x) if self.random_init else torch.zeros_like(x)
        delta = delta.float()
        delta = delta.to(x.device)
        return x, delta, []



from enum import Enum
from typing import Optional, overload

import torch
import requests

from secmlt.models.base_model import BaseModel
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing
from sqlalchemy.util import OrderedDict
from torch.utils.data import DataLoader

from maltorch.utils.utils import compute_hash_from_sample
from maltorch.zoo.model import Model


class VTEngineLabel(Enum):
    AVAST = "avast!"
    AVG = "avg"
    BITDEFENDER = "bitdefender"
    CLAMAV = "clamav"
    CROWDSTRIKE = "crowdstrike"
    DRWEB = "drweb"
    ESET_NOD32 = "eset-nod32"
    F_SECURE = "f-secure"
    KASPERSKY = "kaspersky"
    MALWAREBYTES = "malwarebytes"
    MICROSOFT_DEFENDER = "windows-defender"
    PANDA = "panda"
    SOPHOS = "sophos"
    SYMANTEC_NORTON = "symantec-norton"
    TREND_MICRO = "trendmicro-housecall"
    VIRUSBUSTER = "virusbuster"


class VirusTotalBaseModel(BaseModel):

    def __init__(self,
                 api_key: str,
                 av_list: list[str] = None,
                 threshold: int = 5,
                 preprocessing: DataProcessing = None,
                 postprocessing: DataProcessing = None,
                 **kwargs):
        super().__init__(preprocessing=preprocessing, postprocessing=postprocessing)
        self.api_key = api_key
        self._filter_avs = av_list
        self.threshold = threshold

    def predict(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        y = self._decision_function(x)
        predictions = y.sum(dim=-1) > self.threshold
        return predictions


    def _decision_function(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        y = torch.zeros((x.shape[0], len(VTEngineLabel) if self._filter_avs is None else len(self._filter_avs)))
        for i in range(x.shape[0]):
            hash_xi = compute_hash_from_sample(x[i, :])
            base_url = "https://www.virustotal.com/api/v3/"
            headers = {"x-apikey": self.api_key}
            response = requests.get(f"{base_url}files/{hash_xi}",
                                    headers=headers)
            av_responses = None
            if response.status_code == 200:
                av_responses = response.json()["data"]["scans"]
                if self._filter_avs is not None:
                    av_responses = OrderedDict(
                        [(k, av_responses[k]['detected'] == True) for k in av_responses if k in self._filter_avs])
            else:
                print(f"Error fetching data: {response.status_code}")
            y[i, :] = torch.Tensor(av_responses.values())
        return y

    def gradient(self, x: torch.Tensor, y: int, *args, **kwargs) -> torch.Tensor:
        pass

    def train(self, dataloader: DataLoader) -> "BaseModel":
        pass


class VirusTotal(Model):
    def __init__(self, name: str, api_key: str, gdrive_id: Optional[str]):
        super().__init__(name, None)
        self.api_key = api_key

    def load_pretrained_model(self, device="cpu", model_path=None):
        return None

    @classmethod
    def create_model(
            cls,
            api_key: str,
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
            **kwargs
    ) -> BaseModel:
        ...



import random
import torch
from maltorch.datasets.binary_dataset import BinaryDataset

class FixedSizeChunkDeRandomizedSmoothingDataset(BinaryDataset):
    """
    Daniel Gibert, Giulio Zizzo, Quan Le
    Certified Robustness of Static Deep Learning-based Malware Detectors against Patch and Append Attacks
    AISec'23
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = 2 ** 20,
                 padding_idx: int = 256,
                 min_len: int = None,
                 chunk_size: int = 512,
                 is_training: bool = True):
        super().__init__(
            csv_filepath=csv_filepath,
            goodware_directory=goodware_directory,
            malware_directory=malware_directory,
            max_len=max_len,
            min_len=min_len,
            padding_idx=padding_idx,
        )
        self.chunk_size = chunk_size
        self.is_training = is_training

    def pad_collate_func(self, batch):
        """
        This function splits a tensor of bytes into chunks.
        It works differently at training and at test time.
        During training, given an input example we randomly select a chunk of bytes
        During testing, given an input example we split the example into chunks.
        """
        vecs = []
        labels = []
        if self.is_training is True:  # Select a random chunk from a given executable
            for x, y in batch:
                if x.shape[0] <= self.chunk_size:
                    vecs.append(x)
                else:  # If the executable's size is greater than the specified chunk size
                    start_location = random.randint(0, max(0, x.shape[0] - self.chunk_size))
                    end_location = start_location + self.chunk_size
                    vecs.append(x[start_location: end_location])
                labels.append(y)
            x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
            # stack will give us (B, 1), so index [:,0] to get to just (B)
            y = torch.tensor(labels)
            return x, y
        else:  # Split an executable into chunks
            if len(batch) == 1: # Only implemented for batch sizes equals to 1
                x = batch[0][0]
                x = [x[i:i+self.chunk_size] for i in range(0, len(x), self.chunk_size)]
                x = torch.nn.utils.rnn.pad_sequence(x, batch_first=True, padding_value=self.padding_idx)
                y = batch[0][1]
                return x, y
            else:
                raise NotImplementedError



"""
Learning the PE Header, Malware Detection with Minimal Domain Knowledge
E. Raff, J. Sylvester, C. Nicholas
https://dl.acm.org/doi/10.1145/3128572.3140442
"""
from torch import nn
from maltorch.zoo.model import EmbeddingModel

class LSTMNet(EmbeddingModel):
    def __init__(self,
                embedding_size: int = 16,
                max_len: int = 512,
                min_len: int = 512,
                threshold: float = 0.5,
                padding_idx: int = 256,
                hidden_size: int = 512,
                num_layers: int = 3,
                dropout: float = 0.20
               ):
        super(LSTMNet, self).__init__(
           name="LSTMNet", gdrive_id=None, max_len=max_len, min_len=min_len
        )
        self.seq_len = min_len
        self.threshold = threshold

        # Embedding
        self.embedding_1 = nn.Embedding(
           num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )

        self.lstm = nn.LSTM(
            input_size=embedding_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            batch_first=True,
            bidirectional=True,
            dropout=dropout if num_layers > 1 else 0.0
        )
        self.dropout = nn.Dropout(dropout)
        self.fc = nn.Linear(2*hidden_size, 1)

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def embedding_matrix(self):
        return self.embedding_1.weight

    def _forward_embed_x(self, x):
        """
        x: (B, T) token indices
        lengths: (B,) true lengths before padding
        """
        #emb = self.emb(x)  # (B, T, E)
        x = x.transpose(1, 2)
        lstm_out, _ = self.lstm(x)
        pooled, _ = lstm_out.max(dim=1)  # (B, 2H)
        logits = self.fc(self.dropout(pooled))  # (B, num_classes)
        return logits



import torch
from torch import Tensor

from maltorch.initializers.initializers import ByteBasedInitializer
from maltorch.manipulations.byte_manipulation import ByteManipulation


class ReplacementManipulation(ByteManipulation):
    def __init__(
        self,
        initializer: ByteBasedInitializer,
        domain_constraints=None,
        perturbation_constraints=None,
    ):
        if domain_constraints is None:
            domain_constraints = []
        if perturbation_constraints is None:
            perturbation_constraints = []
        super().__init__(initializer, domain_constraints, perturbation_constraints)

    def _apply_manipulation(
        self, x: torch.Tensor, delta: torch.Tensor
    ) -> tuple[Tensor, Tensor]:
        for i, idx in enumerate(self.indexes_to_perturb):
            idx = idx.to(x.device)
            size_indexes = idx[idx != -1].shape[-1]
            x[i, idx[idx != -1]] = delta[i][:size_indexes].to(x.device).long()
        return x, delta

    def initialize(self, samples: torch.Tensor):
        self.indexes_to_perturb = []
        modified_sample, delta, indexes = self.initializer(samples.data)
        self.indexes_to_perturb = indexes
        samples.data = modified_sample
        return samples, delta



import torch
from secmlt.models.data_processing.data_processing import DataProcessing
import math

class KPartitionDeRandomizedPreprocessing(DataProcessing):
    """
    Shoumik Saha, Weinxio Wang, Yigitcan Kaya, Soheil Feizi, Tudo Dumitras
    DRSM: De-Randomized Smoothing on Malware Classifier Providing Certified Robustness
    ICRL'24
    """
    def __init__(self, num_chunks: int = 4, min_chunk_size: int = 500, padding_idx: int = 256):
        super().__init__()
        self.num_chunks = num_chunks
        self.min_chunk_size = min_chunk_size
        self.padding_idx = padding_idx

    def _process(self, x: torch.Tensor) -> torch.Tensor:
        x = x.squeeze()
        L = len(x)
        chunk_size = math.ceil(L/self.num_chunks)
        if chunk_size < self.min_chunk_size:
            chunk_size = self.min_chunk_size
        x = [x[i:i+chunk_size] for i in range(0, L, chunk_size)]
        x = torch.nn.utils.rnn.pad_sequence(x, batch_first=True, padding_value=self.padding_idx)
        return x

    def invert(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        return x.view(-1)



from pathlib import Path
from typing import Type, Union, List, Callable

import nevergrad
from nevergrad.optimization import Optimizer
from secmlt.models.base_model import BaseModel

from secmlt.trackers import Tracker
from torch.nn import BCEWithLogitsLoss, BCELoss

from maltorch.adv.evasion.base_optim_attack_creator import (
    BaseOptimAttackCreator,
    OptimizerBackends,
)
from maltorch.adv.evasion.gradfree_attack import GradientFreeBackendAttack
from maltorch.initializers.initializers import IdentityInitializer
from maltorch.manipulations.fast_gamma_section_injection_manipulation import FastGAMMASectionInjectionManipulation
from maltorch.manipulations.gamma_section_injection_manipulation import GAMMASectionInjectionManipulation
from maltorch.optim.optimizer_factory import MalwareOptimizerFactory


class GAMMASectionInjectionGradFree(GradientFreeBackendAttack):
    def __init__(
            self,
            query_budget: int,
            benignware_folder: Path,
            how_many_sections: int,
            which_sections: list = None,
            y_target: Union[int, None] = None,
            population_size: int = 10,
            random_init: bool = False,
            model_outputs_logits: bool = True,
            device: str = "cpu",
            reg_parameter: float = 1e-5,
            trackers: Union[List[Tracker], Tracker] = None,
    ):
        if which_sections is None:
            which_sections = ['rodata']
        initializer = IdentityInitializer(
            random_init=random_init
        )
        optimizer_cls = MalwareOptimizerFactory.create_ga(
            population_size=population_size
        )
        loss_function = BCEWithLogitsLoss(reduction="none") if model_outputs_logits else BCELoss(reduction="none")
        manipulation_function = FastGAMMASectionInjectionManipulation(benignware_folder=benignware_folder,
                                                                  which_sections=which_sections,
                                                                  how_many_sections=how_many_sections)
        super().__init__(
            y_target=y_target,
            query_budget=query_budget,
            loss_function=loss_function,
            initializer=initializer,
            manipulation_function=manipulation_function,
            optimizer_cls=optimizer_cls,
            trackers=trackers,
            device=device,
            reg_parameter=reg_parameter
        )

    def _init_optimizer(self, model: BaseModel, delta: nevergrad.p.Array) -> Optimizer:
        self.optimizer = self.optimizer_cls(
            parametrization=nevergrad.p.Array(
                shape=delta.value.shape, lower=0.0, upper=1.0
            ), budget=self.query_budget
        )
        return self.optimizer


class GAMMASectionInjection(BaseOptimAttackCreator):
    """
    GAMMA Section Injection attack

    Demetrio, L., Biggio, B., Lagorio, G., Roli, F., & Armando, A. (2021).
    Functionality-preserving black-box optimization of adversarial windows malware.
    IEEE Transactions on Information Forensics and Security, 16, 3469-3478.
    """

    @staticmethod
    def get_backends() -> set[str]:
        return {OptimizerBackends.NG}

    @staticmethod
    def _get_nevergrad_implementation() -> Type[GAMMASectionInjectionGradFree]:
        return GAMMASectionInjectionGradFree

    def __new__(
            cls,
            query_budget: int,
            benignware_folder: Path,
            which_sections: list = None,
            y_target: Union[int, None] = None,
            how_many_sections: int = 75,
            random_init: bool = False,
            population_size: int = 10,
            reg_parameter: float = 1e-5,
            device: str = "cpu",
            model_outputs_logits: bool = True,
            trackers: Union[List[Tracker], Tracker] = None,
            backend: str = OptimizerBackends.NG,
    ) -> Callable:
        if which_sections is None:
            which_sections = ['rodata']
        return GAMMASectionInjectionGradFree(
            query_budget=query_budget,
            benignware_folder=benignware_folder,
            how_many_sections=how_many_sections,
            which_sections=which_sections,
            population_size=population_size,
            y_target=y_target,
            trackers=trackers,
            random_init=random_init,
            model_outputs_logits=model_outputs_logits,
            device=device,
            reg_parameter=reg_parameter
        )



import torch
from secmlt.optimization.gradient_processing import GradientProcessing


class ByteGradientProcessing(GradientProcessing):
    def __call__(self, grad: torch.Tensor) -> torch.Tensor:
        # grad.data = normalize(grad.data, p=2, dim=2)
        return grad



import abc
from typing import Optional

import lightgbm
from secmlt.models.base_model import BaseModel
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing

from maltorch.data_processing.thrember_preprocessing import THREMBERPreprocessing
from maltorch.zoo.gbdt import GBDTModel
from maltorch.zoo.model import Model



class ThremberGBDT(Model):
    def __init__(self, model_path: Optional[str] = None):
        # https://drive.google.com/file/d/1UUYwTT9ST_0Cs_854aVHXUOfgSVXACaD/view?usp=drive_link
        super().__init__(
            name="ember2024_gbdt", gdrive_id="1UUYwTT9ST_0Cs_854aVHXUOfgSVXACaD"
        )
        self.tree_model = None
        self.load_pretrained_model(model_path=model_path)

    def load_pretrained_model(self, device: str = "cpu", model_path: str = None):
        if model_path is None:
            self._fetch_pretrained_model()
            self.tree_model = lightgbm.Booster(model_file=self.model_path)
        else:
            self.tree_model = lightgbm.Booster(model_file=model_path)

    @classmethod
    def create_model(
            cls,
            model_path: Optional[str] = None,
            threshold: int = 0.5,
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
    ) -> BaseModel:
        model = cls(model_path=model_path)
        return GBDTModel(tree_model=model.tree_model, threshold=threshold, preprocessing=THREMBERPreprocessing())



import torch
from secmlt.models.base_model import BaseModel

from maltorch.adv.evasion.backend_attack import BackendAttack
from maltorch.optim.base import BaseByteOptimizer


class GradientBackendAttack(BackendAttack):
    def _init_attack_manipulation(
        self, samples: torch.Tensor
    ) -> (torch.Tensor, torch.Tensor):
        x_adv, delta = self.manipulation_function.initialize(samples.data)
        x_adv = x_adv.long()
        delta = delta.to(samples.device)
        delta.requires_grad = True
        return x_adv, delta

    def _apply_manipulation(
        self, x: torch.Tensor, delta: torch.Tensor
    ) -> (torch.Tensor, torch.Tensor):
        x.data, delta.data = self.manipulation_function(x.data, delta.data)
        return x, delta

    def _init_optimizer(
        self, model: BaseModel, delta: torch.Tensor
    ) -> BaseByteOptimizer:
        self.optimizer = self.optimizer_cls(
            [delta],
            indexes_to_perturb=self.manipulation_function.indexes_to_perturb,
            model=model,
        )
        return self.optimizer

    def _optimizer_step(self, delta: torch.Tensor, loss: torch.Tensor) -> torch.Tensor:
        loss.sum().backward()
        self.optimizer.step()
        return delta

    def _consumed_budget(self):
        return 1



"""
Coull, Scott E., and Christopher Gardner.
"Activation analysis of a byte-based deep neural network for malware classification."
2019 IEEE Security and Privacy Workshops (SPW). IEEE, 2019.
"""
from typing import Optional, Union

import torch
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing

from maltorch.data_processing.e2e_preprocessor import PaddingPreprocessing
from maltorch.zoo.model import EmbeddingModel, BaseEmbeddingPytorchClassifier


class BBDnn(EmbeddingModel):
    DEFAULT_MAX_LENGTH = 102400
    DEFAULT_MIN_LENGTH = 4096

    def __init__(
            self,
            embedding_size: int = 10,
            min_len: int = DEFAULT_MIN_LENGTH,
            max_len: int = DEFAULT_MAX_LENGTH,
            threshold: float = 0.5,
            padding_idx: int = 256,
    ):
        super(BBDnn, self).__init__(name="bbdnn", gdrive_id="1c_9lVHT9zYpBCwQfnUW6ZbCF6SaVabRZ", min_len=min_len,
                                    max_len=max_len)
        self.max_len = max_len
        self.threshold = threshold
        self.embedding_1 = torch.nn.Embedding(
            num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx
        )
        self.conv1d_1 = torch.nn.Conv1d(
            in_channels=embedding_size,
            out_channels=96,
            kernel_size=(11,),
            stride=(1,),
            groups=1,
            bias=True,
        )
        self.conv1d_2 = torch.nn.Conv1d(
            in_channels=96,
            out_channels=128,
            kernel_size=(5,),
            stride=(1,),
            groups=1,
            bias=True,
        )
        self.conv1d_3 = torch.nn.Conv1d(
            in_channels=128,
            out_channels=256,
            kernel_size=(5,),
            stride=(1,),
            groups=1,
            bias=True,
        )
        self.conv1d_4 = torch.nn.Conv1d(
            in_channels=256,
            out_channels=256,
            kernel_size=(5,),
            stride=(1,),
            groups=1,
            bias=True,
        )
        self.conv1d_5 = torch.nn.Conv1d(
            in_channels=256,
            out_channels=256,
            kernel_size=(5,),
            stride=(1,),
            groups=1,
            bias=True,
        )
        self.dense_1 = torch.nn.Linear(in_features=512, out_features=1, bias=True)

    def embedding_layer(self):
        return self.embedding_1

    def _forward_embed_x(self, x):
        conv1d_1 = self.conv1d_1(x)
        conv1d_1_activation = torch.relu(conv1d_1)
        max_pooling1d_1 = torch.max_pool1d(
            conv1d_1_activation,
            kernel_size=(4,),
            stride=(4,),
            padding=0,
            ceil_mode=False,
        )

        conv1d_2 = self.conv1d_2(max_pooling1d_1)
        conv1d_2_activation = torch.relu(conv1d_2)
        max_pooling1d_2 = torch.max_pool1d(
            conv1d_2_activation,
            kernel_size=(4,),
            stride=(4,),
            padding=0,
            ceil_mode=False,
        )

        conv1d_3 = self.conv1d_3(max_pooling1d_2)
        conv1d_3_activation = torch.relu(conv1d_3)
        max_pooling1d_3 = torch.max_pool1d(
            conv1d_3_activation,
            kernel_size=(4,),
            stride=(4,),
            padding=0,
            ceil_mode=False,
        )

        conv1d_4 = self.conv1d_4(max_pooling1d_3)
        conv1d_4_activation = torch.relu(conv1d_4)
        max_pooling1d_4 = torch.max_pool1d(
            conv1d_4_activation,
            kernel_size=(4,),
            stride=(4,),
            padding=0,
            ceil_mode=False,
        )

        conv1d_5 = self.conv1d_5(max_pooling1d_4)
        conv1d_5_activation = torch.relu(conv1d_5)
        global_max_pooling1d_1 = torch.max_pool1d(
            input=conv1d_5_activation, kernel_size=conv1d_5_activation.size()[2:]
        )
        global_average_pooling1d_1 = torch.avg_pool1d(
            input=conv1d_5_activation, kernel_size=conv1d_5_activation.size()[2:]
        )
        global_max_pooling1d_1_flatten = global_max_pooling1d_1.view(
            global_max_pooling1d_1.size(0), -1
        )
        global_average_pooling1d_1_flatten = global_average_pooling1d_1.view(
            global_average_pooling1d_1.size(0), -1
        )
        concatenate_1 = torch.cat(
            (global_max_pooling1d_1_flatten, global_average_pooling1d_1_flatten), 1
        )
        dense_1 = self.dense_1(concatenate_1)
        return dense_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def embedding_matrix(self):
        return self.embedding_1.weight

    @classmethod
    def create_model(
            cls,
            model_path: Optional[str] = None,
            device: str = "cpu",
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
            threshold: Optional[Union[float, None]] = 0.5,
            **kwargs,
    ) -> BaseEmbeddingPytorchClassifier:
        if preprocessing is None:
            preprocessing = PaddingPreprocessing(max_len=BBDnn.DEFAULT_MAX_LENGTH)
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net = net.to(device)  # Explicitly load model to device
        net = net.eval()
        net = BaseEmbeddingPytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
            threshold=threshold,
        )
        return net



import lightgbm
import torch
from secmlt.models.base_model import BaseModel
from secmlt.models.data_processing.data_processing import DataProcessing
from torch.utils.data import DataLoader


class GBDTModel(BaseModel):
    def __init__(self, tree_model: lightgbm.Booster, threshold: float = 0.82, preprocessing: DataProcessing = None,
                 postprocessing: DataProcessing = None):
        super().__init__(preprocessing=preprocessing, postprocessing=postprocessing)
        self.tree_model: lightgbm.Booster = tree_model
        self.threshold: float = threshold

    def predict(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        y = self.decision_function(x)
        return torch.Tensor(y > self.threshold).int()

    def _decision_function(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        probabilities = self.tree_model.predict(x.numpy())
        probabilities = torch.Tensor(probabilities).unsqueeze(1)
        return probabilities

    def gradient(self, x: torch.Tensor, y: int, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("No gradient for GBDT model.")

    def train(self, dataloader: DataLoader) -> "BaseModel":
        raise NotImplementedError("Trainer not implemented yet.")



from typing import Tuple
from PIL import Image
from torch.utils.data import Dataset
from torchvision.transforms import Resize, Compose, Normalize, PILToTensor, ConvertImageDtype, Grayscale
import torch


def get_size(bytez_length: int) -> Tuple[int, int]:
    if (bytez_length < 10240):
        width = 32
    elif (10240 <= bytez_length <= 10240 * 3):
        width = 64
    elif (10240 * 3 <= bytez_length <= 10240 * 6):
        width = 128
    elif (10240 * 6 <= bytez_length <= 10240 * 10):
        width = 256
    elif (10240 * 10 <= bytez_length <= 10240 * 20):
        width = 384
    elif (10240 * 20 <= bytez_length <= 10240 * 50):
        width = 512
    elif (10240 * 50 <= bytez_length <= 10240 * 100):
        width = 768
    else:
        width = 1024

    height = int(bytez_length / width) + 1
    return width, height

class GrayscaleDataset(Dataset):
    def __init__(self, csv_filepath, width=256, height=256, convert_to_3d_image=True):
        self.all_files = []
        self.width = width
        self.height = height
        self.convert_to_3d_image = convert_to_3d_image
        if self.convert_to_3d_image:
            self.preprocess = Compose(
                [
                    PILToTensor(),
                    ConvertImageDtype(torch.float),
                    Resize((self.width, self.height)),
                    Normalize((0.5), (0.5)),
                    Grayscale(num_output_channels=3)
                ]
            )
        else:
            self.preprocess = Compose(
                [
                    PILToTensor(),
                    ConvertImageDtype(torch.float),
                    Resize((self.width, self.height)),
                    Normalize((0.5), (0.5))
                ]
            )

        with open(csv_filepath, "r") as input_file:
            lines = input_file.readlines()
            for line in lines:
                tokens = line.strip().split(",")
                self.all_files.append([tokens[0], int(tokens[1])])

    def __len__(self) -> int:
        return len(self.all_files)

    def __getitem__(self, index) -> Tuple[torch.Tensor, torch.Tensor]:
        to_load, label = self.all_files[index]
        img = self.convert_to_grayscale_img(to_load)
        return img, torch.tensor(label)

    def convert_to_grayscale_img(self, to_load) -> torch.Tensor:
        with open(to_load, "rb") as input_file:
            bytez = input_file.read()
        width, height = get_size(len(bytez))

        image = Image.new("L", (width,height))
        image.putdata(bytez)
        image = self.preprocess(image)
        image = torch.cat([image, image, image])

        return image




import random
import string
from pathlib import Path
from typing import Union

import lief.PE
import torch

from maltorch.initializers.initializers import IdentityInitializer
from maltorch.manipulations.byte_manipulation import ByteManipulation
from maltorch.utils.utils import convert_torch_exe_to_list
from maltorch.utils.zangope import Binary


class FastGAMMASectionInjectionManipulation(ByteManipulation):
    def __init__(
            self,
            benignware_folder: Union[Path, list[str], list[Path]],
            which_sections=None,
            how_many_sections: int = 75,
            domain_constraints=None,
            perturbation_constraints=None,
    ):
        self.how_many_sections = how_many_sections
        if which_sections is None:
            which_sections = [".rdata"]
        self.benignware_folder = benignware_folder
        self.which_sections = which_sections
        if domain_constraints is None:
            domain_constraints = []
        if perturbation_constraints is None:
            perturbation_constraints = []
        self._sections = []
        self._names = [
            ''.join(random.choices(string.ascii_uppercase + string.digits, k=8)) for _ in range(self.how_many_sections)
        ]
        for path in sorted(self.benignware_folder.glob("*")):
            if not lief.is_pe(str(path)):
                continue
            lief_pe = lief.parse(str(path))
            for s in lief_pe.sections:
                if s.name not in self.which_sections:
                    continue
                if len(self._sections) < self.how_many_sections:
                    self._sections.append(list(s.content))
        super().__init__(IdentityInitializer(), domain_constraints, perturbation_constraints)

    def _apply_manipulation(
            self, x: torch.Tensor, delta: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        pe = Binary(bytez=bytearray(convert_torch_exe_to_list(x)))
        for delta_i, content, name in zip(delta.squeeze(), self._sections, self._names):
            pe.add_robust_section(name, 0x40000040, bytearray(content[:int(len(content) * delta_i)]))
        x = torch.atleast_2d(torch.Tensor(pe.get_bytes()).long()).to(x.device)
        return x, delta

    def initialize(self, samples: torch.Tensor):
        delta = torch.zeros((samples.shape[0], self.how_many_sections))
        if self.initializer.random_init:
            delta = torch.rand((samples.shape[0], self.how_many_sections))
        return samples, delta


from typing import Type, Union, List, Callable

from secmlt.trackers import Tracker
from torch.nn import BCEWithLogitsLoss, BCELoss

from maltorch.adv.evasion.base_optim_attack_creator import (
    BaseOptimAttackCreator,
    OptimizerBackends,
)
from maltorch.adv.evasion.gradfree_attack import GradientFreeBackendAttack
from maltorch.adv.evasion.gradient_attack import GradientBackendAttack
from maltorch.initializers.section_injection_initializer import SectionInjectionInitializer
from maltorch.manipulations.replacement_manipulation import ReplacementManipulation
from maltorch.optim.optimizer_factory import MalwareOptimizerFactory


class SectionInjectionGradFree(GradientFreeBackendAttack):
    def __init__(
            self,
            query_budget: int,
            how_many_sections: int,
            size_per_section: int,
            y_target: Union[int, None] = None,
            population_size: int = 10,
            random_init: bool = False,
            model_outputs_logits: bool = True,
            device:str="cpu",
            trackers: Union[List[Tracker], Tracker] = None,
    ):
        initializer = SectionInjectionInitializer(
            random_init=random_init, how_many_sections=how_many_sections, size_per_section=size_per_section
        )
        optimizer_cls = MalwareOptimizerFactory.create_ga(
            population_size=population_size
        )
        loss_function = BCEWithLogitsLoss(reduction="none") if model_outputs_logits else BCELoss(reduction="none")
        manipulation_function = ReplacementManipulation(initializer=initializer)
        super().__init__(
            y_target=y_target,
            query_budget=query_budget,
            loss_function=loss_function,
            initializer=initializer,
            manipulation_function=manipulation_function,
            optimizer_cls=optimizer_cls,
            trackers=trackers,
            device=device
        )


class SectionInjectionGrad(GradientBackendAttack):
    def __init__(
            self,
            query_budget: int,
            how_many_sections: int,
            size_per_section: int,
            y_target: Union[int, None] = None,
            random_init: bool = False,
            step_size: int = 58,
            device: str = "cpu",
            model_outputs_logits: bool = True,
            trackers: Union[List[Tracker], Tracker] = None,
    ):
        initializer = SectionInjectionInitializer(
            random_init=random_init, how_many_sections=how_many_sections, size_per_section=size_per_section
        )
        optimizer_cls = MalwareOptimizerFactory.create_bgd(
            lr=step_size,
            device=device
        )
        loss_function = BCEWithLogitsLoss(reduction="none") if model_outputs_logits else BCELoss(reduction="none")
        manipulation_function = ReplacementManipulation(initializer=initializer)
        super().__init__(
            y_target=y_target,
            query_budget=query_budget,
            loss_function=loss_function,
            optimizer_cls=optimizer_cls,
            manipulation_function=manipulation_function,
            initializer=initializer,
            trackers=trackers,
            model_outputs_logits=model_outputs_logits,
            device=device
        )


class SectionInjection(BaseOptimAttackCreator):
    """
    Section Injection attack

    Demetrio, L., Biggio, B., Lagorio, G., Roli, F., & Armando, A. (2021).
    Functionality-preserving black-box optimization of adversarial windows malware.
    IEEE Transactions on Information Forensics and Security, 16, 3469-3478.
    """

    @staticmethod
    def get_backends() -> set[str]:
        return {OptimizerBackends.GRADIENT, OptimizerBackends.NG}

    @staticmethod
    def _get_nevergrad_implementation() -> Type[SectionInjectionGradFree]:
        return SectionInjectionGradFree

    @staticmethod
    def _get_native_implementation() -> Type[SectionInjectionGrad]:
        return SectionInjectionGrad

    def __new__(
            cls,
            query_budget: int,
            y_target: Union[int, None] = None,
            how_many_sections: int = 75,
            size_per_section: int = 512,
            random_init: bool = False,
            step_size: int = 16,
            population_size: int = 10,
            device: str = "cpu",
            model_outputs_logits: bool = True,
            trackers: Union[List[Tracker], Tracker] = None,
            backend: str = OptimizerBackends.GRADIENT,
    ) -> Callable:
        implementation: Callable = cls.get_implementation(backend)
        if backend == OptimizerBackends.GRADIENT:
            kwargs = {"step_size": step_size}
        else:
            kwargs = {
                "population_size": population_size,
            }
        return implementation(
            query_budget=query_budget,
            how_many_sections=how_many_sections,
            size_per_section=size_per_section,
            y_target=y_target,
            trackers=trackers,
            random_init=random_init,
            model_outputs_logits=model_outputs_logits,
            device=device,
            **kwargs,
        )



PE_HEADER_OFFSET = 60
PE_HEADER_OFFSET_LENGTH = 4

PE_MAGIC_LENGTH = 4
OPT_HEADER_MAGIC_LENGTH = 2
COFF_LENGTH = 20
COFF_N_SECTION = 2

DLL_CHARACTERISTICS_OFFSET_FROM_PE = 70
DLL_CHARACTERISTICS_LENGTH = 2

SECTION_EXECUTABLE = 0x20000000
SECTION_READABLE = 0x40000000
SECTION_WRITEABLE = 0x80000000

DYNAMIC_REBASE = 0x0040
FORCE_INTEGRITY = 0x0080

SECTION_ENTRY_LENGTH = 40

PE_32_MAGIC = 0x10b
PE_64_MAGIC = 0x20b


def align(value, align_on):
    r = value % align_on
    if r > 0:
        return value + (align_on - r)
    return value


class Binary:
    def __init__(self, path=None, bytez=None):
        if path is None and bytez is None:
            raise ValueError("You must pass as input either a path or bytes")
        self.exe_bytes = bytearray()
        if path is not None:
            with open(path, 'rb') as f:
                self.exe_bytes = bytearray(f.read())
        else:
            self.exe_bytes = bytearray(bytez)

    @classmethod
    def load_from_path(cls, path: str):
        return cls(path=path)

    @classmethod
    def load_from_bytes(cls, bytez: bytearray):
        return cls(bytez=bytez)

    @staticmethod
    def flag_is_on(characteristics, flag):
        return (characteristics & flag) == flag

    @staticmethod
    def deactivate_flag(characteristics, flag):
        if Binary.flag_is_on(characteristics, flag):
            characteristics &= ~flag
        return characteristics

    @staticmethod
    def activate_flag(characteristics, flag):
        if not Binary.flag_is_on(characteristics, flag):
            characteristics |= flag
        return characteristics

    def get_bytes(self):
        return self.exe_bytes

    def get_pe_location(self):
        pe_location = self.exe_bytes[PE_HEADER_OFFSET:PE_HEADER_OFFSET + PE_HEADER_OFFSET_LENGTH]
        return int.from_bytes(pe_location, 'little')

    def get_optional_header_location(self):
        pe_location = self.get_pe_location()
        optional_header_location = pe_location + PE_MAGIC_LENGTH + 20
        return optional_header_location

    def get_optional_header_size(self):
        pe_location = self.get_pe_location()
        size_opt_header_location = pe_location + PE_MAGIC_LENGTH + 16
        size_opt_header = self.exe_bytes[size_opt_header_location: size_opt_header_location + 2]
        size_opt_header = int.from_bytes(size_opt_header, 'little')
        return size_opt_header

    def get_section_table_location(self):
        size_opt_header = self.get_optional_header_size()
        return self.get_optional_header_location() + size_opt_header

    def get_total_number_sections(self):
        pe_location = self.get_pe_location()
        n_sections = self.exe_bytes[
                     pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION: pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION + 2]
        n_sections = int.from_bytes(n_sections, 'little')
        return n_sections

    def increase_number_sections(self):
        pe_location = self.get_pe_location()
        n_sections = self.exe_bytes[
                     pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION: pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION + 2]
        n_sections = int.from_bytes(n_sections, 'little') + 1
        self.exe_bytes[
        pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION: pe_location + PE_MAGIC_LENGTH + COFF_N_SECTION + 2] = n_sections.to_bytes(
            2, 'little')

    def get_section_entry_location_from_index(self, index: int):
        n_sections = self.get_total_number_sections()
        if index > n_sections:
            raise ValueError(f"Section with index {index} not found. Only {n_sections} are present.")
        section_table_offset = self.get_section_table_location()
        return section_table_offset + index * 40

    def get_section_entry_from_index(self, index: int):
        n_sections = self.get_total_number_sections()
        if index > n_sections:
            raise ValueError(f"Section with index {index} not found. Only {n_sections} are present.")
        section_table_offset = self.get_section_table_location()
        return self.exe_bytes[section_table_offset + index * 40: section_table_offset + (index + 1) * 40]

    def get_section_alignment(self):
        optional_header_location = self.get_optional_header_location()
        section_alignment = int.from_bytes(
            self.exe_bytes[optional_header_location + 32: optional_header_location + 32 + 4], 'little')
        return section_alignment

    def get_file_alignment(self):
        optional_header_location = self.get_optional_header_location()
        file_alignment = int.from_bytes(
            self.exe_bytes[optional_header_location + 36: optional_header_location + 36 + 4], 'little')
        return file_alignment

    def get_sizeof_headers(self):
        optional_header_location = self.get_optional_header_location()
        sizeof_headers = int.from_bytes(
            self.exe_bytes[optional_header_location + 60: optional_header_location + 60 + 4], 'little')
        return sizeof_headers  # multiple of FileAlignment

    def set_sizeof_headers(self, value: int):
        optional_header_location = self.get_optional_header_location()
        self.exe_bytes[optional_header_location + 60: optional_header_location + 60 + 4:] = value.to_bytes(4, 'little')
        return value  # multiple of FileAlignment

    def get_sizeof_image(self):
        optional_header_location = self.get_optional_header_location()
        sizeof_image = int.from_bytes(self.exe_bytes[optional_header_location + 56: optional_header_location + 56 + 4],
                                      'little')
        return sizeof_image  # multiple of SectionAlignment

    def set_sizeof_image(self, value: int):
        optional_header_location = self.get_optional_header_location()
        self.exe_bytes[optional_header_location + 56: optional_header_location + 56 + 4:] = value.to_bytes(4, 'little')

    def increase_pointer_raw_section(self, section_index: int, value: int):
        section_table_offset = self.get_section_table_location() + section_index * 40
        old_pointer = int.from_bytes(self.exe_bytes[section_table_offset + 20: section_table_offset + 20 + 4], 'little')
        new_pointer = old_pointer + value
        self.exe_bytes[section_table_offset + 20: section_table_offset + 20 + 4] = new_pointer.to_bytes(4, 'little')


    def rva_to_file_offset(self, rva: int):
        # Helper: convert an RVA to file offset using section table
        # iterate sections and find which contains rva
        total = self.get_total_number_sections()
        st = self.get_section_table_location()
        for i in range(total):
            entry = self.exe_bytes[st + i * SECTION_ENTRY_LENGTH: st + (i + 1) * SECTION_ENTRY_LENGTH]
            va = int.from_bytes(entry[12:16], 'little')
            vsz = int.from_bytes(entry[8:12], 'little')
            ptr = int.from_bytes(entry[20:24], 'little')
            # NOTE: VirtualSize may be zero, but SizeOfRawData can still be >0; use VirtualSize if present else SizeOfRawData
            if vsz == 0:
                vsz = int.from_bytes(entry[16:20], 'little')
            if va <= rva < (va + vsz):
                return ptr + (rva - va)
        return None

    def add_robust_section(self, name: str, characteristics: int, content: bytearray):
        n_sections = self.get_total_number_sections()
        if n_sections == 0:
            raise RuntimeError("Binary has no sections; cannot add a new section safely.")

        # Read last section header
        last_section = self.get_section_entry_from_index(n_sections - 1)
        last_virtual_address = int.from_bytes(last_section[12:16], 'little')
        last_virtual_size = int.from_bytes(last_section[8:12], 'little')

        section_alignment = self.get_section_alignment()
        file_alignment = self.get_file_alignment()

        # compute new section VA correctly
        aligned_last_virtual_size = align(last_virtual_size, section_alignment)
        next_virtual_address = last_virtual_address + aligned_last_virtual_size

        virtual_size = len(content)
        size_of_raw_data = align(virtual_size, file_alignment)

        # compose IMAGE_SECTION_HEADER (40 bytes)
        new_section_entry = bytearray(SECTION_ENTRY_LENGTH)
        name_bytes = name.encode('ascii', errors='ignore')[:8]
        new_section_entry[0:len(name_bytes)] = name_bytes
        new_section_entry[8:12] = (virtual_size).to_bytes(4, 'little')  # VirtualSize
        new_section_entry[12:16] = next_virtual_address.to_bytes(4, 'little')  # VirtualAddress
        new_section_entry[16:20] = size_of_raw_data.to_bytes(4, 'little')  # SizeOfRawData
        # PointerToRawData will be set later (we choose EOF aligned to file_alignment)
        new_section_entry[36:40] = characteristics.to_bytes(4, 'little')  # Characteristics

        # where the new section header should go in the section table
        section_table_offset = self.get_section_table_location()
        new_section_header_offset = section_table_offset + n_sections * SECTION_ENTRY_LENGTH

        sizeof_headers = self.get_sizeof_headers()

        # locate DataDirectory base and relevant entries (cert & debug)
        opt_loc = self.get_optional_header_location()
        opt_magic = int.from_bytes(self.exe_bytes[opt_loc: opt_loc + 2], 'little')
        # data directories start offset relative to Optional Header:
        # PE32 -> offset 96, PE32+ -> offset 112
        if opt_magic == PE_32_MAGIC:
            data_dir_base = opt_loc + 96
        else:
            data_dir_base = opt_loc + 112
        # certificate directory entry (IMAGE_DIRECTORY_ENTRY_SECURITY) is index 4, each entry 8 bytes
        cert_dir_offset = data_dir_base + 4 * 8
        # debug directory entry is index 6 (IMAGE_DIRECTORY_ENTRY_DEBUG)
        debug_dir_offset = data_dir_base + 6 * 8

        # If the new section header doesn't fit in SizeOfHeaders, expand headers
        if new_section_header_offset + SECTION_ENTRY_LENGTH > sizeof_headers:
            # compute minimal increment to make room, align to FileAlignment
            needed = (new_section_header_offset + SECTION_ENTRY_LENGTH) - sizeof_headers
            increment = align(needed, file_alignment)

            insert_at = new_section_header_offset  # we insert padding at the end of the current section table
            # insert zeros into file bytes (this shifts everything after insert_at)
            self.exe_bytes = self.exe_bytes[:insert_at] + (b'\x00' * increment) + self.exe_bytes[insert_at:]

            # update SizeOfHeaders
            self.set_sizeof_headers(sizeof_headers + increment)

            # shift all existing section PointerToRawData by increment
            for i in range(n_sections):
                self.increase_pointer_raw_section(i, increment)

            # Certificate directory: VirtualAddress field actually stores a file offset (per PE spec).
            try:
                cert_va = int.from_bytes(self.exe_bytes[cert_dir_offset: cert_dir_offset + 4], 'little')
                cert_size = int.from_bytes(self.exe_bytes[cert_dir_offset + 4: cert_dir_offset + 8], 'little')
                if cert_va != 0 and cert_va >= insert_at:
                    new_cert_va = cert_va + increment
                    self.exe_bytes[cert_dir_offset: cert_dir_offset + 4] = new_cert_va.to_bytes(4, 'little')
            except IndexError:
                # optional header smaller/invalid — ignore safely
                pass

            # Debug directory: if present, it is referenced by a data-directory RVA (index 6)
            try:
                debug_rva = int.from_bytes(self.exe_bytes[debug_dir_offset: debug_dir_offset + 4], 'little')
                debug_size = int.from_bytes(self.exe_bytes[debug_dir_offset + 4: debug_dir_offset + 8], 'little')
                if debug_rva != 0 and debug_size != 0:
                    debug_file_offset = self.rva_to_file_offset(debug_rva)
                    if debug_file_offset is not None:
                        # IMAGE_DEBUG_DIRECTORY is 28 bytes each
                        entry_size = 28
                        count = debug_size // entry_size
                        for di in range(count):
                            entry_off = debug_file_offset + di * entry_size
                            # PointerToRawData is at offset 24 (4 bytes)
                            ptr_field_off = entry_off + 24
                            # Defensive bounds check
                            if ptr_field_off + 4 <= len(self.exe_bytes):
                                old_ptr = int.from_bytes(self.exe_bytes[ptr_field_off: ptr_field_off + 4], 'little')
                                if old_ptr != 0 and old_ptr >= insert_at:
                                    new_ptr = old_ptr + increment
                                    self.exe_bytes[ptr_field_off: ptr_field_off + 4] = new_ptr.to_bytes(4, 'little')
            except (IndexError, ValueError):
                # Not present or malformed - ignore
                pass

            # update sizeof_headers var for subsequent checks
            sizeof_headers = self.get_sizeof_headers()

        # After potential expansion, choose PointerToRawData: append at EOF aligned to FileAlignment
        pointer_to_raw = align(len(self.exe_bytes), file_alignment)
        new_section_entry[20:24] = pointer_to_raw.to_bytes(4, 'little')

        # final safety check: the new header must now fit in SizeOfHeaders
        if new_section_header_offset + SECTION_ENTRY_LENGTH > sizeof_headers:
            raise RuntimeError("Failed to make room for the new section header inside SizeOfHeaders; aborting.")

        # write the new section header
        self.exe_bytes[new_section_header_offset: new_section_header_offset + SECTION_ENTRY_LENGTH] = new_section_entry

        # append raw data at EOF (ensure space)
        needed_len = pointer_to_raw + size_of_raw_data
        if len(self.exe_bytes) < needed_len:
            self.exe_bytes += b'\x00' * (needed_len - len(self.exe_bytes))

        # write content
        self.exe_bytes[pointer_to_raw: pointer_to_raw + len(content)] = content

        # update SizeOfImage properly
        new_sizeof_image = next_virtual_address + align(virtual_size, section_alignment)
        new_sizeof_image = align(new_sizeof_image, section_alignment)
        self.set_sizeof_image(new_sizeof_image)

        # increase NumberOfSections
        self.increase_number_sections()

        return self.exe_bytes



"""
Malware Detection by Eating a Whole EXE
Edward Raff, Jon Barker, Jared Sylvester, Robert Brandon, Bryan Catanzaro, Charles Nicholas
https://arxiv.org/abs/1710.09435
"""
from typing import Optional, Union

import torch
import torch.nn.functional as F
from secmlt.models.base_trainer import BaseTrainer
from secmlt.models.data_processing.data_processing import DataProcessing
from torch import nn

from maltorch.data_processing.e2e_preprocessor import PaddingPreprocessing
from maltorch.zoo.model import EmbeddingModel, BaseEmbeddingPytorchClassifier


class MalConv(EmbeddingModel):
    """
    Architecture implementation.
    """
    DEFAULT_MAX_LENGTH = 2**20

    def __init__(self, embedding_size: int = 8,
                 max_len: int =DEFAULT_MAX_LENGTH,
                 threshold: float =0.5,
                 padding_idx: int = 256,
                 kernel_size=512,
                 stride=512):
        #https://drive.google.com/file/d/1Uk7QHjjXMEy-RADX5kHD9vIYk6UT2nii/view?usp=drive_link
        super(MalConv, self).__init__(
            name="MalConv", gdrive_id="1Uk7QHjjXMEy-RADX5kHD9vIYk6UT2nii", max_len=max_len
        )
        out_channels = 128
        output_size=1
        self.embedding_1 = nn.Embedding(num_embeddings=257, embedding_dim=embedding_size, padding_idx=padding_idx)
        self.conv1d_1 = nn.Conv1d(in_channels=embedding_size, out_channels=out_channels, kernel_size=(kernel_size,),
                                  stride=(stride,),
                                  groups=1, bias=True)
        self.conv1d_2 = nn.Conv1d(in_channels=embedding_size, out_channels=out_channels, kernel_size=(kernel_size,),
                                  stride=(stride,),
                                  groups=1, bias=True)
        self.dense_1 = nn.Linear(in_features=out_channels, out_features=out_channels, bias=True)
        self.dense_2 = nn.Linear(in_features=out_channels, out_features=output_size, bias=True)
        self.embedding_size = (embedding_size,)
        self.max_len = max_len
        self.threshold = threshold
        self.invalid_value = padding_idx

    def embedding_layer(self):
        return self.embedding_1

    def embed(self, x):
        emb_x = self.embedding_1(x)
        emb_x = emb_x.transpose(1, 2)
        return emb_x

    def _forward_embed_x(self, x):
        conv1d_1 = self.conv1d_1(x)
        conv1d_2 = self.conv1d_2(x)
        conv1d_1_activation = torch.relu(conv1d_1)
        conv1d_2_activation = torch.sigmoid(conv1d_2)
        multiply_1 = conv1d_1_activation * conv1d_2_activation
        global_max_pooling1d_1 = F.max_pool1d(
            input=multiply_1, kernel_size=multiply_1.size()[2:]
        )
        global_max_pooling1d_1_flatten = global_max_pooling1d_1.view(
            global_max_pooling1d_1.size(0), -1
        )
        dense_1 = self.dense_1(global_max_pooling1d_1_flatten)
        dense_1_activation = torch.relu(dense_1)
        dense_2 = self.dense_2(dense_1_activation)
        return dense_2

    def embedding_matrix(self):
        return self.embedding_1.weight

    @classmethod
    def create_model(
            cls,
            model_path: Optional[str] = None,
            device: str = "cpu",
            preprocessing: DataProcessing = None,
            postprocessing: DataProcessing = None,
            trainer: BaseTrainer = None,
            threshold: Optional[Union[float, None]] = 0.5,
            **kwargs,
    ) -> BaseEmbeddingPytorchClassifier:
        if preprocessing is None:
            preprocessing = PaddingPreprocessing(max_len=MalConv.DEFAULT_MAX_LENGTH)
        net = cls(**kwargs)
        net.load_pretrained_model(device=device, model_path=model_path)
        net = net.to(device)  # Explicitly load model to device
        net = net.eval()
        net = BaseEmbeddingPytorchClassifier(
            model=net,
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            trainer=trainer,
            threshold=threshold,
        )
        return net



from pathlib import Path
from typing import Optional, Union

import torch


def load_from_folder(
        path: Path, extension: Optional[str] = None, padding: int = 256, limit=None, device="cpu"
) -> torch.Tensor:
    """Create a torch.Tensor whose rows are all the file with extension specified in input.
    Tensor are padded to match the same size.
    :param path: Folder path
    :param extension: default "exe", filters all the file based on this extension
    :param padding: default 256, pad every tensor with this value to uniform the size
    :param limit: default None, limit the number of loaded file, None for load all folder
    :return: a torch.Tensor containing all the file converted into tensors
    """
    X = []
    if extension is None:
        pattern = "*"
    else:
        pattern = f"*.{extension}"
    for filepath in sorted(path.glob(pattern)):
        x = load_single_exe(filepath)
        if x is not None:
            X.append(x.float())
            if limit is not None and len(X) >= limit:
                break
    X = torch.nn.utils.rnn.pad_sequence(X, padding_value=padding).transpose(0, 1).long()
    X = X.to(device)
    return X


def create_labels(x: torch.Tensor, label: int, device="cpu"):
    """
    Create the labels for the specified data.
    """
    y = torch.zeros((x.shape[0], 1)) + label
    y = y.to(device)
    return y


def load_single_exe(path: Path) -> Union[torch.Tensor, None]:
    """
    Create a torch.Tensor from the file pointed in the path
    :param path: a pathlib Path
    :return: torch.Tensor containing the bytes of the file as a tensor, None if the file is not an exe
    """
    with open(path, "rb") as h:
        code = h.read()
    if len(code) != 0 and code[:2] == b'MZ':
        return torch.frombuffer(bytearray(code), dtype=torch.uint8).to(torch.long)
    return None



from typing import Union, List, Type, Callable

from secmlt.trackers.trackers import Tracker
from torch.nn import BCEWithLogitsLoss, BCELoss

from maltorch.adv.evasion.base_optim_attack_creator import (
    BaseOptimAttackCreator,
    OptimizerBackends,
)
from maltorch.adv.evasion.gradfree_attack import GradientFreeBackendAttack
from maltorch.adv.evasion.gradient_attack import GradientBackendAttack
from maltorch.manipulations.gradient_free_wrapper_manipulation import GradientFreeWrapperManipulation
from maltorch.manipulations.replacement_manipulation import (
    ReplacementManipulation,
)
from maltorch.initializers.dos_initializer import PartialDOSInitializer
from maltorch.optim.optimizer_factory import MalwareOptimizerFactory


class PartialDOSGradFree(GradientFreeBackendAttack):
    def __init__(
            self,
            query_budget: int,
            y_target: Union[int, None] = None,
            population_size: int = 10,
            random_init: bool = False,
            model_outputs_logits: bool = True,
            device: str = "cpu",
            trackers: Union[List[Tracker], Tracker] = None,
    ):
        loss_function = BCEWithLogitsLoss(reduction="none") if model_outputs_logits else BCELoss(reduction="none")
        initializer = PartialDOSInitializer(random_init=random_init)
        manipulation_function = GradientFreeWrapperManipulation(ReplacementManipulation(initializer=initializer))
        optimizer_cls = MalwareOptimizerFactory.create_ga(
            population_size=population_size
        )
        super().__init__(
            y_target=y_target,
            query_budget=query_budget,
            loss_function=loss_function,
            optimizer_cls=optimizer_cls,
            manipulation_function=manipulation_function,
            initializer=initializer,
            model_outputs_logits=model_outputs_logits,
            trackers=trackers,
            device=device
        )


class PartialDOSGrad(GradientBackendAttack):
    def __init__(
            self,
            query_budget: int,
            y_target: Union[int, None] = None,
            random_init: bool = False,
            step_size: int = 58,
            device: str = "cpu",
            model_outputs_logits: bool = True,
            trackers: Union[List[Tracker], Tracker] = None,
    ):
        loss_function = BCEWithLogitsLoss(reduction="none") if model_outputs_logits else BCELoss(reduction="none")
        initializer = PartialDOSInitializer(random_init=random_init)
        manipulation_function = ReplacementManipulation(initializer=initializer)
        optimizer_cls = MalwareOptimizerFactory.create_bgd(lr=step_size, device=device)
        super().__init__(
            y_target=y_target,
            query_budget=query_budget,
            loss_function=loss_function,
            optimizer_cls=optimizer_cls,
            manipulation_function=manipulation_function,
            initializer=initializer,
            model_outputs_logits=model_outputs_logits,
            trackers=trackers,
            device=device
        )


class PartialDOS(BaseOptimAttackCreator):
    @staticmethod
    def get_backends() -> set[str]:
        return {OptimizerBackends.GRADIENT, OptimizerBackends.NG}

    @staticmethod
    def _get_nevergrad_implementation() -> Type[PartialDOSGradFree]:
        return PartialDOSGradFree

    @staticmethod
    def _get_native_implementation() -> Type[PartialDOSGrad]:
        return PartialDOSGrad

    def __new__(
            cls,
            query_budget: int,
            y_target: Union[int, None] = None,
            random_init: bool = False,
            step_size: int = 58,
            population_size: int = 10,
            device: str = "cpu",
            model_outputs_logits: bool = True,
            trackers: Union[List[Tracker], Tracker] = None,
            backend: str = OptimizerBackends.GRADIENT,
    ) -> Callable:
        implementation: Callable = cls.get_implementation(backend)
        if backend == OptimizerBackends.GRADIENT:
            kwargs = {"step_size": step_size}
        else:
            kwargs = {
                "population_size": population_size,
            }
        return implementation(
            query_budget=query_budget,
            y_target=y_target,
            trackers=trackers,
            random_init=random_init,
            model_outputs_logits=model_outputs_logits,
            device=device,
            **kwargs,
        )



import torch
from secmlt.models.data_processing.data_processing import DataProcessing


class PaddingPreprocessing(DataProcessing):
    """
    The preprocessor needed to pad each test-time sample to the maximum length of the networks.
    """
    def _process(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        padding = torch.nn.ConstantPad1d((0, self.max_len - x.shape[-1]), 256)
        return padding(x)

    def invert(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        pass

    def __init__(self, max_len:int=2**20):
        self.max_len = max_len


import torch
from secmlt.models.data_processing.data_processing import DataProcessing
from torch.nn.functional import sigmoid


class MajorityVotingPostprocessing(DataProcessing):
    def __init__(self, threshold: float = 0.5, apply_sigmoid:bool=False):
        super().__init__()
        self.threshold = threshold
        self.apply_sigmoid = apply_sigmoid

    def _process(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        if self.apply_sigmoid:
            x = sigmoid(x)
        x = x.squeeze()
        y_preds = torch.where(x >= self.threshold, 1, 0)
        num_benign = torch.sum(y_preds == 0)
        num_malicious = torch.sum(y_preds == 1)
        return torch.atleast_2d(num_malicious / (num_benign + num_malicious))

    def invert(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
        Apply the inverted transform (if defined).

        Parameters
        ----------
        x : torch.Tensor
            Input samples.

        Returns
        -------
        torch.Tensor
            The samples in the input space before the transformation.
        """
        raise NotImplementedError



import torch
import lief
from maltorch.initializers.initializers import ByteBasedInitializer
from maltorch.utils.utils import convert_torch_exe_to_list


class PartialDOSInitializer(ByteBasedInitializer):
    def __init__(self, random_init: bool = False):
        super().__init__(random_init)

    def __call__(self, x: torch.Tensor):
        indexes = torch.arange(2, 60).long().repeat((x.shape[0], 1))
        delta = (
            x.data[:, 2:60]
            if not self.random_init
            else torch.randint(0, 255, (x.shape[0], 58))
        ).float()
        return x, delta, indexes


class DOSHeaderStubInitializer(ByteBasedInitializer):
    def __init__(self, random_init: bool = False):
        super().__init__(random_init)

    def __call__(self, x: torch.Tensor):
        indexes = []
        for i in range(x.shape[0]):
            lief_x =  lief.PE.parse(convert_torch_exe_to_list(x[i]))
            pe_position = lief_x.dos_header.addressof_new_exeheader
            x_indexes = list(range(2, 60)) + list(range(64, pe_position))
            indexes.append(torch.LongTensor(x_indexes))
        indexes = torch.nn.utils.rnn.pad_sequence(indexes, padding_value=-1).transpose(0, 1).long()
        delta = (
            torch.zeros_like(indexes)
            if not self.random_init
            else torch.randint(0, 255, indexes.shape)
        ).float()
        delta[indexes == -1] = -1
        delta = delta.float()
        delta = delta.to(x.device)
        indexes = indexes.to(x.device)
        return x, delta, indexes


"""
Trizna, D. and Demetrio, L. and Battista, B. and Fabio, R. (2024)
Nebula: Self-Attention for Dynamic Malware Analysis
IEEE Transactions on Information Forensics and Security (TIFS)
Paper: https://ieeexplore.ieee.org/document/10551436
Source: https://github.com/dtrizna/nebula/blob/main/nebula/preprocessing/pe.py
"""

import os
import json
import logging
from time import time
from pathlib import Path
from typing import Union

import speakeasy
from maltorch.utils.strings import get_alphanum_chars

from .constants import (
    SPEAKEASY_RECORD_FIELDS,
    SPEAKEASY_RECORD_LIMITS
)
from .tokenization import JSONFilter
from .normalization import normalize_table_ip, normalize_table_path


class PEDynamicFeatureExtractor:
    def __init__(
        self,
        speakeasy_config=None,
        speakeasy_record_fields=SPEAKEASY_RECORD_FIELDS,
        record_limits=SPEAKEASY_RECORD_LIMITS,
        emulation_output_folder=None
    ):
        # Setup Speakeasy config
        if speakeasy_config is None:
            speakeasy_config = os.path.join(os.path.dirname(__file__), "speakeasy_config.json")
        if isinstance(speakeasy_config, dict):
            self.speakeasy_config = speakeasy_config
        else:
            if not os.path.exists(speakeasy_config):
                raise FileNotFoundError(f"Speakeasy config file not found: {speakeasy_config}")
            with open(speakeasy_config, "r") as f:
                self.speakeasy_config = json.load(f)

        self.record_limits = record_limits
        self.parser = JSONFilter(fields=speakeasy_record_fields)

        self.output_folder = emulation_output_folder
        if self.output_folder:
            os.makedirs(self.output_folder, exist_ok=True)

    def _create_error_file(self, errfile):
        # Just creating an empty file to indicate failure
        Path(errfile).touch()

    def _emulation(self, config: dict, data: bytes, path: str = None):
        try:
            file_name = path if path else str(data[:15])
            se = speakeasy.Speakeasy(config=config)
            if path:
                module = se.load_module(path=path)
            elif data:
                module = se.load_module(data=data)
            else:
                raise ValueError("[-] Either 'path' or 'data' must be provided to load_module.")
            se.run_module(module)
            return se.get_report()
        except Exception as ex:
            logging.error(f"[-] Failed emulation of {file_name} | Exception:\n{ex}")
            return None

    def emulate(self, raw_pe: Union[str, bytes]):
        if isinstance(raw_pe, str):
            if os.path.exists(raw_pe):
                with open(raw_pe, "rb") as f:
                    raw_pe = f.read()
                sample_name = os.path.splitext(os.path.basename(raw_pe))[0]
            else:
                raise FileNotFoundError(f"[-] File not found: {raw_pe}")
        else:
            sample_name = f"{int(time())}"
        report = self._emulation(self.speakeasy_config, data=raw_pe, path=sample_name)

        if self.output_folder:
            if report:
                output_path = os.path.join(self.output_folder, f"{self.sample_name}.json")
                with open(output_path, "w") as f:
                    json.dump(report, f, indent=4)
            else:
                err_file = os.path.join(self.output_folder, f"{self.sample_name}.err")
                self._create_error_file(err_file)

        if report and 'entry_points' in report:
            api_seq_len = sum(len(entry.get("apis", [])) for entry in report["entry_points"])
        else:
            api_seq_len = 0
        return (
            self.filter_and_normalize_report(report["entry_points"])
            if api_seq_len > 0 else None
        )

    def filter_and_normalize_report(self, entry_points):
        if isinstance(entry_points, str) and os.path.exists(entry_points):
            with open(entry_points, "r") as f:
                entry_points = json.load(f)
        # Clean up report
        record_dict = self.parser.filter_and_concat(entry_points)

        # Filter out events with uninformative API sequences
        if (
            'apis' in record_dict and
            record_dict['apis'].shape[0] == 1 and
            record_dict['apis'].iloc[0].api_name == 'MSVBVM60.ordinal_100'
        ):
            return None

        # Normalize data
        if 'file_access' in record_dict:
            record_dict['file_access'] = normalize_table_path(
                record_dict['file_access'], col='path'
            )
        if (
            'network_events.traffic' in record_dict and
            'server' in record_dict['network_events.traffic'].columns
        ):
            record_dict['network_events.traffic'] = normalize_table_ip(
                record_dict['network_events.traffic'], col='server'
            )
        if (
            'network_events.dns' in record_dict and
            'query' in record_dict['network_events.dns'].columns
        ):
            record_dict['network_events.dns']['query'] = (
                record_dict['network_events.dns']['query']
                .apply(lambda x: ' '.join(x.split('.')))
            )
        # Normalize args to exclude any non-alphanumeric characters
        if 'args' in record_dict['apis'].columns:
            record_dict['apis']['args'] = record_dict['apis']['args'].apply(
                lambda args_list: [get_alphanum_chars(arg) for arg in args_list]
            )

        # Limit verbose fields to a certain number of records
        if self.record_limits:
            for field, limit in self.record_limits.items():
                if field in record_dict:
                    record_dict[field] = record_dict[field].head(limit)
        # Join records
        record_json = self.join_records_to_json(record_dict)
        return record_json

    @staticmethod
    def join_records_to_json(record_dict):
        # Sort keys to ensure consistent order, put 'apis' at the end
        sorted_keys = sorted(record_dict.keys(), reverse=True)
        json_event_parts = [
            f"\"{key}\":{record_dict[key].to_json(orient='records')}"
            for key in sorted_keys
        ]
        json_event = "{" + ",".join(json_event_parts) + "}"
        return json.loads(json_event)



from secmlt.models.data_processing.data_processing import DataProcessing
import torch
import ember


class EMBERPreprocessing(DataProcessing):
    """
    Anderson, H. S., & Roth, P. (2018).
    EMBER: an open dataset for training static pe malware machine learning models. 
    arXiv preprint arXiv:1804.04637.
    """

    def invert(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        raise NotImplementedError("Not supported.")

    def _process(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        extractor = ember.features.PEFeatureExtractor(print_feature_warning=False)
        n_samples = x.shape[0]
        feat_x = torch.zeros((x.shape[0], extractor.dim))
        for i in range(n_samples):
            x_i = x[i, :]
            x_i = x_i.type(torch.int).flatten().tolist()
            if 256 in x_i:
                x_i = x_i[:x_i.index(256)]
            x_bytes = bytearray(x_i)
            feat_x[i, :] = torch.Tensor(extractor.feature_vector(x_bytes))
        return feat_x



"""PyTorch model trainers with early stopping."""

import torch.nn
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader
from tqdm import tqdm
import copy


class EarlyStoppingPyTorchTrainer:
    """Trainer for PyTorch models with early stopping."""

    def __init__(self, optimizer: torch.optim.Optimizer, epochs: int = 5,
                 loss: torch.nn.Module = None, scheduler: _LRScheduler = None) -> None:
        """
        Create PyTorch trainer.
        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to use for training the model.
        epochs : int, optional
            Number of epochs, by default 5.
        loss : torch.nn.Module, optional
            Loss to minimize, by default None.
        scheduler : _LRScheduler, optional
            Scheduler for the optimizer, by default None.
        """
        self._epochs = epochs
        self._optimizer = optimizer
        self._loss = loss if loss is not None else torch.nn.BCEWithLogitsLoss()
        self._scheduler = scheduler

        self.training_losses = []
        self.training_accuracies = []
        self.validation_losses = []
        self.validation_accuracies = []

    def train(self, model: torch.nn.Module,
            train_loader: DataLoader,
            val_loader: DataLoader,
            patience: int) -> torch.nn.Module:
        """
        Train model with given loaders and early stopping.
        Parameters
        ----------
        model : torch.nn.Module
            Pytorch model to be trained.
        train_loader : DataLoader
            Train data loader.
        val_loader : DataLoader
            Validation data loader.
        patience : int
            Number of epochs to wait before early stopping.
        Returns
        -------
        torch.nn.Module
            Trained model.
        """
        # Check model has .threshold
        if not hasattr(model, 'threshold'):
            raise AttributeError("Model must have a 'threshold' attribute for binary classification.")

        best_loss = float("inf")
        best_model = None
        patience_counter = 0
        for epoch in range(self._epochs):
            model = self.fit(model, train_loader)
            val_loss = self.validate(model, val_loader)
            if val_loss <= best_loss:
                best_loss = val_loss
                best_model = copy.deepcopy(model)
                best_model = best_model.to(next(model.parameters()).device)
                patience_counter = 0
            else:
                patience_counter += 1
            if patience_counter >= patience:
                print(f"Early stopping triggered. The validation losss hasn't improved for {patience_counter} epochs")
                break
            print(
                f"Epoch {epoch}: val_loss = {val_loss}, best_loss = {best_loss}, patience_counter = {patience_counter}")
        return best_model

    def fit(self,
              model: torch.nn.Module,
              dataloader: DataLoader) -> torch.nn.Module:
        """
        Train model for one epoch with given loader.
        Parameters
        ----------
        model : torch.nn.Module
            Pytorch model to be trained.
        dataloader : DataLoader
            Train data loader.
        Returns
        -------
        torch.nn.Module
            Trained model.
        """
        device = next(model.parameters()).device
        model = model.train()
        running_loss = 0.0
        train_total = 0
        train_correct = 0
        num_batches = 0
        for x, y in tqdm(dataloader):
            x, y = x.to(device), y.to(device)
            self._optimizer.zero_grad()
            outputs = model(x)
            outputs = outputs.view(-1)
            loss = self._loss(outputs, y.float())
            loss.backward()
            self._optimizer.step()

            running_loss += loss.item()
            probs = torch.sigmoid(outputs)
            y_preds = (probs >= model.threshold).int()

            train_total += y.size(0)
            train_correct += (y_preds == y).sum().item()
            num_batches += 1
        self.training_losses.append(running_loss / num_batches)
        self.training_accuracies.append(train_correct / train_total)

        if self._scheduler is not None:
            self._scheduler.step()
        return model

    def validate(self,
                 model: torch.nn.Module,
                 dataloader: DataLoader) -> float:
        """
        Validate model with given loader.
        Parameters
        ----------
        model : torch.nn.Module
            Pytorch model to be validated.
        dataloader : DataLoader
            Validation data loader.
        Returns
        -------
        float
            Validation loss of the model.
        """
        running_loss = 0
        val_total = 0
        val_correct = 0
        device = next(model.parameters()).device
        model = model.eval()
        num_batches = 0
        with torch.no_grad():
            for x, y in tqdm(dataloader):
                x, y = x.to(device), y.to(device)
                outputs = model(x)
                outputs = outputs.view(-1)
                loss = self._loss(outputs, y.float())
                running_loss += loss.item()
                probs = torch.sigmoid(outputs)
                y_preds = (probs >= model.threshold).int()
                val_total += y.size(0)
                val_correct += (y_preds == y).sum().item()
                num_batches += 1

            val_loss = running_loss / num_batches
            self.validation_losses.append(running_loss / num_batches)
            self.validation_accuracies.append(val_correct / val_total)
        return val_loss



from pathlib import Path

import pandas as pd
from torch.utils.data import Dataset
from typing import Tuple, Union
import torch
import os



class BinaryDataset(Dataset):
    """
    Load samples from sources.
    """
    def __init__(self,
                 csv_filepath: str = None,
                 goodware_directory: str = None,
                 malware_directory: str = None,
                 max_len: int = None,
                 padding_idx: int = 256,
                 min_len: int = None,
                 max_date: str = None,
                 min_date: str = None):
        self.all_files = []
        self.max_len = max_len
        self.min_len = min_len
        self.padding_idx = padding_idx

        if csv_filepath is not None and max_date is None and min_date is None:
            with open(csv_filepath, "r") as input_file:
                lines = input_file.readlines()
                for line in lines:
                    tokens = line.strip().split(",")
                    self.all_files.append([tokens[0], int(tokens[1]), os.path.getsize(tokens[0])])


        # in this case, csv file must contain a column labeled "timestamp" formatted as YYYY-MM, and it will be read with pandas.
        # The rest of the file should have column labels "path", "hash", "label"
        elif csv_filepath is not None:
            rows = pd.read_csv(csv_filepath)

            # Ensure required columns are present
            required_columns = {"timestamp", "hash", "path", "label"}
            if not required_columns.issubset(rows.columns):
                raise ValueError(f"CSV file must contain the following columns: {required_columns}")
            if min_date is not None and max_date is None:
                rows = rows[
                    (pd.to_datetime(rows["timestamp"], format="%Y-%m") >= pd.to_datetime(min_date, format="%Y-%m"))]
            # if two data boundaries are provided, the data will be filtered to only include samples within the range
            elif min_date is not None and max_date is not None:
                rows = rows[
                    (pd.to_datetime(rows["timestamp"], format="%Y-%m") >= pd.to_datetime(min_date, format="%Y-%m")) &
                    (pd.to_datetime(rows["timestamp"], format="%Y-%m") <= pd.to_datetime(max_date, format="%Y-%m"))
                ]
            elif max_date is not None and min_date is None:
                rows = rows[pd.to_datetime(rows["timestamp"], format="%Y-%m") <= pd.to_datetime(max_date, format="%Y-%m")]
            else:
                raise ValueError("max_date or/and min_date must be specified to filter the dataset by timestamp")
            for _, row in rows.iterrows():
                self.all_files.append([os.path.join(row["path"], row["hash"]), row["label"],
                                       os.path.getsize(os.path.join(row["path"], row["hash"]))])

        elif goodware_directory is not None or malware_directory is not None:
            if goodware_directory is not None:
                self.all_files.extend(
                    [[os.path.join(goodware_directory, filename), 0, os.path.getsize(os.path.join(goodware_directory, filename))] for filename in os.listdir(goodware_directory)])
            if malware_directory is not None:
                self.all_files.extend(
                    [[os.path.join(malware_directory, filename), 1, os.path.getsize(os.path.join(malware_directory, filename))] for filename in os.listdir(malware_directory)])
        else:
            raise NotImplementedError("You need to either provide CSV file containing (sample,label) "
                                      "or the paths where the goodware and malware are stored.")

    def __len__(self) -> int:
        return len(self.all_files)

    def __getitem__(self, index) -> Tuple[torch.Tensor, torch.Tensor]:
        to_load, label, _ = self.all_files[index]
        x = self.load_and_pad_samples(to_load, max_len=self.max_len, min_len=self.min_len, padding_idx=self.padding_idx)
        return x, torch.tensor(label)

    def pad_collate_func(self, batch):
        """
        This should be used as the collate_fn=pad_collate_func for a pytorch DataLoader object in order to pad out files in a batch to the length of the longest item in the batch.
        """
        vecs = [x[0] for x in batch]
        labels = [x[1].unsqueeze(0) for x in batch]

        x = torch.nn.utils.rnn.pad_sequence(vecs, batch_first=True, padding_value=self.padding_idx)
        # stack will give us (B, 1), so index [:,0] to get to just (B)
        y = torch.stack(labels).float()

        return x, y

    def load_and_pad_samples(self, path: Path, max_len:int, min_len:int, padding_idx:int=256) -> Union[torch.Tensor, None]:
        """
        Create a torch.Tensor from the file pointed in the path
        :param path: a pathlib Path
        :return: torch.Tensor containing the bytes of the file as a tensor, None if the file is not an exe
        """
        with open(path, "rb") as h:
            code = h.read(max_len)
        x = torch.frombuffer(bytearray(code), dtype=torch.uint8).to(torch.float)
        if len(code) < min_len:
            padding = torch.nn.ConstantPad1d((0, len(code) - min_len), padding_idx)
            x = padding(x)
        return x



import torch

from maltorch.initializers.initializers import ByteBasedInitializer
from maltorch.utils.pe_operations import extend_manipulation


class ExtendDOSInitializer(ByteBasedInitializer):
    def __init__(self, preferred_manipulation_size, random_init=False):
        super().__init__(random_init=random_init)
        self.manipulation_size = preferred_manipulation_size

    def __call__(self, x: torch.Tensor) -> [torch.Tensor, torch.Tensor]:
        X = []
        deltas = []
        indexes = []
        for x_i in x:
            x_i, shift_indexes = extend_manipulation(x_i, self.manipulation_size)
            X.append(x_i)
            delta = (
                torch.zeros(len(shift_indexes))
                if not self.random_init
                else torch.randint(0, 255, (len(shift_indexes),))
            )
            deltas.append(delta)
            indexes.append(torch.Tensor(shift_indexes))
        device = x.device
        x, delta, indexes = self._pad_samples_same_length(X, deltas, indexes)
        x = x.long()
        x = x.to(device)
        delta = delta.float()
        delta = delta.to(device)
        indexes = indexes.to(device)
        return x, delta, indexes



from typing import Callable, Optional

import torch.optim
from torch import Tensor

from maltorch.optim.base import BaseByteOptimizer

INVALID = torch.inf


class BGD(BaseByteOptimizer):
    def step(self, closure: Optional[Callable[[], float]] = ...) -> Optional[float]:
        for group in self.param_groups:
            for delta in group["params"]:
                if self._embedding_grad is None:
                    raise ValueError(
                        "Grad is none, you should call backward first before invoking the step function. "
                        "Otherwise, the model itself is not differentiable and gradients can not be computed."
                    )
                grad_embedding = self._embedding_grad[0]
                # grad_embedding = self.gradient_processing(grad_embedding)
                updated_delta = token_gradient_descent(
                    embedding_tokens=self.embedding_matrix,
                    emb_x=self.embedding_matrix[delta.long()],
                    gradient_f=grad_embedding,
                    step_size=self.step_size,
                    index_to_perturb=self.indexes_to_perturb,
                    unavailable_tokens=torch.LongTensor([256]),
                )
                updated_delta = updated_delta.to(delta.device)
                to_update = updated_delta != INVALID
                if torch.any(to_update):
                    delta.data[to_update] = updated_delta[to_update]
        return None


def token_gradient_descent(
    embedding_tokens: torch.Tensor,
    emb_x: torch.Tensor,
    gradient_f: torch.Tensor,
    step_size: int,
    index_to_perturb: torch.LongTensor,
    unavailable_tokens: Optional[torch.Tensor] = None,
) -> Tensor:
    optimized_tokens = torch.zeros(*emb_x.shape[:2]) + INVALID
    for j in range(emb_x.shape[0]):
        step_size_index = (
            gradient_f[j, :][index_to_perturb[j].view(-1)]
            .norm(dim=1)
            .argsort(descending=True)[:step_size]
        )
        for i in step_size_index:
            byte_in_grad = index_to_perturb[j, i]
            gradient_f_i = -gradient_f[j, byte_in_grad] #maltorch is written to minimize the loss
            x_i = emb_x[j, i]
            token_to_chose = single_token_gradient_update(
                start_token=x_i,
                gradient=gradient_f_i,
                embedded_tokens=embedding_tokens,
                unavailable_tokens=unavailable_tokens,
            )
            optimized_tokens[j, i] = token_to_chose
    return optimized_tokens


def single_token_gradient_update(
    start_token: torch.Tensor,
    gradient: torch.Tensor,
    embedded_tokens: torch.Tensor,
    invalid_val=INVALID,
    unavailable_tokens: Optional[torch.Tensor] = None,
):
    """
    Given the starting byte, the gradient and the embedding map,it returns a list of distances

    Parameters
    ----------
    start_token : torch.Tensor
        the starting embedding token for the search
    gradient : torch.Tensor
        the gradient of a single embedded token
    embedded_tokens : torch.Tensor
        the embedding matrix with all the byte embedded
    admitted_tokens: list
        the list of indexes of the tokens to use in the search
    invalid_val : optional, default torch.inf
        the invalid value to use. Default torch.inf
    unavailable_tokens: Union[torch.Tensor, None] = None
        if specified, it avoids the usage of the selected tokens during the search step
    Returns
    -------

    """
    if torch.equal(gradient, torch.zeros_like(gradient)):
        return invalid_val
    gradient = gradient / gradient.norm()
    B = embedded_tokens
    distance = torch.full((B.shape[0],), invalid_val, device=B.device)
    same_mask = torch.all(B == start_token, dim=1)
    if unavailable_tokens is not None:
        unavail_mask = torch.zeros_like(distance, dtype=torch.bool)
        unavail_mask[unavailable_tokens] = True
    else:
        unavail_mask = torch.zeros_like(distance, dtype=torch.bool)

    W = B - start_token  # (N, D)
    s = torch.sum(W * gradient, dim=1)  # (N,)
    forward_mask = s >= 0
    valid_mask = (~same_mask) & (~unavail_mask) & forward_mask
    if not torch.any(valid_mask):
        return invalid_val

    proj = start_token + s[valid_mask].unsqueeze(1) * gradient  # (Nv, D)
    dist_valid = torch.norm(B[valid_mask] - proj, dim=1)
    distance[valid_mask] = dist_valid
    min_value, token_to_choose = torch.min(distance, dim=0)

    if min_value == invalid_val:
        return invalid_val

    return token_to_choose



import torch

from maltorch.initializers.initializers import ByteBasedInitializer
from maltorch.utils.pe_operations import padding_manipulation


class PaddingInitializer(ByteBasedInitializer):
    def __init__(self, padding: int = 2048, random_init: bool = False):
        super().__init__(random_init)
        self.padding = padding

    def __call__(self, x: torch.Tensor):
        X = []
        deltas = []
        indexes = []
        for x_i in x:
            x_i, padding_indexes = padding_manipulation(x_i.unsqueeze(0), self.padding)
            X.append(x_i.squeeze())
            delta = (
                torch.zeros(len(padding_indexes))
                if not self.random_init
                else torch.randint(0, 255, (len(padding_indexes),))
            )
            deltas.append(delta)
            indexes.append(padding_indexes)
        device = x.device
        x, delta, indexes = self._pad_samples_same_length(X, deltas, torch.Tensor(indexes))
        x = x.long()
        x = x.to(device)
        delta = delta.float()
        delta = delta.to(device)
        indexes = indexes.to(device)
        return x, delta, indexes



import nevergrad
import numpy as np
import torch
from nevergrad.optimization import Optimizer
from secmlt.models.base_model import BaseModel
from torch.utils.data import DataLoader, TensorDataset

from maltorch.adv.evasion.backend_attack import BackendAttack


class GradientFreeBackendAttack(BackendAttack):
    def _optimizer_step(
        self, delta: nevergrad.p.Array, loss: torch.Tensor
    ) -> nevergrad.p.Array:
        self.optimizer.tell(delta, loss.item())
        return self.optimizer.ask()

    def _apply_manipulation(
        self, x: torch.Tensor, delta: nevergrad.p.Array
    ) -> (torch.Tensor, torch.Tensor):
        p_delta = torch.from_numpy(delta.value)
        p_delta = p_delta.float()
        p_delta = p_delta.to(x.device)
        x, torch_delta = self.manipulation_function(x.data, p_delta)
        return x, delta

    def _consumed_budget(self):
        return 1

    def _init_attack_manipulation(
        self, samples: torch.Tensor
    ) -> (torch.Tensor, nevergrad.p.Array):
        x_adv, delta = super()._init_attack_manipulation(samples)
        optim_delta = nevergrad.p.Array(shape=delta.shape, lower=0.0, upper=1.0)
        optim_delta.value = delta.cpu().numpy()
        return x_adv, optim_delta

    def _init_optimizer(self, model: BaseModel, delta: nevergrad.p.Array) -> Optimizer:
        self.optimizer = self.optimizer_cls(
            parametrization=nevergrad.p.Array(
                shape=delta.value.shape, lower=0.0, upper=1.0
            ), budget=self.query_budget
        )
        return self.optimizer

    def _delta_norm(self, x: nevergrad.p.Array):
        return np.linalg.norm(x.value, ord=1)

    def __call__(self, model: BaseModel, data_loader: DataLoader) -> DataLoader:
        adversarials = []
        original_labels = []
        for samples, labels in data_loader:
            for sample, label in zip(samples, labels):
                sample = sample.unsqueeze(0)
                label = label.unsqueeze(0)
                x_adv, _ = self._run(model, sample, label)
                adversarials.append(x_adv.transpose(0,1))
                original_labels.append(label)
        adversarials = (
            torch.nn.utils.rnn.pad_sequence(adversarials, padding_value=256)
            .transpose(0,1)
            .squeeze()
            .long()
        )
        adversarials = torch.atleast_2d(adversarials)
        original_labels = torch.vstack(original_labels)
        adversarial_dataset = TensorDataset(adversarials, original_labels)
        return DataLoader(
            adversarial_dataset,
            batch_size=data_loader.batch_size,
        )

    def _init_best_tracking(self, delta: torch.Tensor): ...

    def _track_best(self, loss: torch.Tensor, delta: torch.Tensor): ...

    def _get_best_delta(self):
        return self.optimizer.provide_recommendation()



import torch
from secmlt.models.data_processing.data_processing import DataProcessing


class SigmoidPostprocessor(DataProcessing):
    def _process(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        return torch.sigmoid(x)

    def invert(self, x: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        pass